#lang typed/racket


(define-type r:value r:number)
(define-type r:expression (U r:value r:apply r:variable))
(struct r:number ([value : Real]))
(struct r:variable ([name : Symbol]))
(struct r:apply ([func : r:expression] [arg1 : r:expression] [arg2 : r:expression] ))

(: r:eval-builtin (-> Symbol (-> Real Real (Option Real))))
(define (r:eval-builtin sym)
  (match sym
    ['+ +]
    ['* *]
    ['/ (lambda (n d) (match d [0 #f] [_ (/ n d)]))]
  )
)
(: handle-err
  (All [Input Output]
    (->
      (Option Input)
      (-> Input (Option Output))
      (Option Output)
    )
  )
)
(define (handle-err res kont)
  (cond
    [(false? res) #f]
    [else (kont res)]))

#|
(define-syntax do
  (syntax-rules (<-)
    ; Only one monadic-op, return it
    [(_ mexp) mexp]
    ; A binding operation
    [(_ var : ty <- mexp rest ...) (handle-err mexp (lambda ([var : ty]) (do rest ...)))]
    ; No binding operator, just ignore the return value
    [(_ mexp rest ...)        (handle-err mexp (lambda (_) (do rest ...)))]))|#

;; (: handle-err
;;   (All [Input Result] (-> (Option Input) (-> Input Result)) (Option Result)))
;; (define (handle-err v)
;;   (cond
;;     [(false? v) #f]
;;     [else (kont v)]
;;   )
;; )

(: r:eval-exp (-> r:expression (Option Real)))
(define (r:eval-exp exp)
  (match exp
    ; If it's a number, return that number
    [(r:number v) v]
    ; If it's a function with 2 arguments
    [(r:apply (r:variable f) arg1 arg2)
      (define func (r:eval-builtin f))
      (handle-err (r:eval-exp arg1)
        (lambda ([a1 : Real])
          (handle-err (r:eval-exp arg2)
            (lambda ([a2 : Real])
              (func a1 a2)
            )
          )
        )
      )
    ]
  )
)
;; (r:eval-exp
;;   (r:apply (r:variable '/)
;;     (r:number 10)
;;     (r:number 2)
;;   )
;; )
;; (r:eval-exp
;;   (r:apply (r:variable '/)
;;     (r:number 10)
;;     (r:number 0)
;;   )
;; )
(: f (-> Real (-> Real Void) Void))
(define (f x ret)
  (ret (+ x 2)))

;(f 10 displayln)

#|
; Version 1
(: r:eval-exp (-> r:expression (Option Real)))
(define (r:eval-exp exp)
  (match exp
    ; If it's a number, return that number
    [(r:number v) v]
    ; If it's a function with 2 arguments
    [(r:apply (r:variable f) arg1 arg2)
      (define func (r:eval-builtin f))
      (match* ((r:eval-exp arg1) (r:eval-exp arg2))
        [((? real? x) (? real? y))
          (func x y)
        ]
        [(_ _) #f]
      )
    ]
  )
)
(r:eval-exp
  (r:apply (r:variable '/)
    (r:number 10)
    (r:number 0)
  )
)
|#
#|
; Version 2

(: r:eval-exp (-> r:expression (Option Real)))
(define (r:eval-exp exp)
  (match exp
    ; If it's a number, return that number
    [(r:number v) v]
    ; If it's a function with 2 arguments
    [(r:apply (r:variable f) arg1 arg2)
      (define func (r:eval-builtin f))
      (define a1 (r:eval-exp arg1))
      (cond [(false? a1) #f]
            [else
              (define a2 (r:eval-exp arg2))
              (cond [(false? a2) #f]
                [else
                  (func a1 a2)])])]))
|#
#|
; Version 3

(: r:eval-exp (-> r:expression (Option Real)))
(define (r:eval-exp exp)
  (match exp
    ; If it's a number, return that number
    [(r:number v) v]
    ; If it's a function with 2 arguments
    [(r:apply (r:variable f) arg1 arg2)
      (define func (r:eval-builtin f))
      (handle-err (r:eval-exp arg1)
        (lambda ([a1 : Real])
          (handle-err (r:eval-exp arg2)
            (lambda ([a2 : Real]) : (Option Real)
              (func a1 a2)
            )
          )
        )
      )
    ]
  )
)
|#
#|
; Version 4
(define (r:eval-exp exp)
  (match exp
    ; If it's a number, return that number
    [(r:number v) v]
    ; If it's a function with 2 arguments
    [(r:apply (r:variable f) arg1 arg2)
      (define func (r:eval-builtin f))
      (do
        a1 : Real <- (r:eval-exp arg1)
        a2 : Real <- (r:eval-exp arg2)
        (func a1 a2)
      )
    ]
  )
)
|#

(define-type (CPS Ok Error Result)
  (->
    (-> Ok Result)
    (-> Error Result)
    Result
  )
)

; `Returns x via the return function`
(: return (All (Ok Error Result)
  (-> Ok (CPS Ok Error Result))))
(define (return x)
  (lambda ([ret : (-> Ok Result)] [err : (-> Error Result)] )
    (ret x)))
; `Returns x via the error function`
(: raise (All (Ok Error Result)
  (-> Error (CPS Ok Error Result))))
(define (raise x)
  (lambda (ret err)
    (err x)))
; `Monadic-bind on CPS-style code`
(: cps-bind
  (All [Ok1 Ok2 Error Result]
    (->
      ; The type of the first continuation
      (CPS Ok1 Error Result)
      ; Given the result of the first, return a continuation
      (-> Ok1 (CPS Ok2 Error Result))
      ; The second continuation
      (CPS Ok2 Error Result)
    )
  )
)
(define (cps-bind o1 o2)
  (lambda ([ret : (-> Ok2 Result)] [err : (-> Error Result)])
    ; Run the first operation
    (o1
      (lambda ([res : Ok1])
        ; If the operation is successful
        (
          ; Build the second CPS operation
          (o2 res)
          ; And pass the original ret/err pairs
          ret
          err
        )
      )
      ; Otherwise, pass the same error
      err
    )
  )
)
;; ; `The try-catch operation`
(: try
  (All [Ok Error1 Error2 Result]
    (->
      ; The type of the first continuation
      (CPS Ok Error1 Result)
      ; Given the result of the first, return a continuation
      (-> Error1 (CPS Ok Error2 Result))
      ; The second continuation
      (CPS Ok Error2 Result)
    )
  )
)
(define (try o1 o2)
  (lambda ([ret : (-> Ok Result)] [err : (-> Error2 Result)])
    ; Try to run o1
    (o1
      ; If successfull, pass it to ret
      ret
      ; Otherwise, call the `catch` block
      (lambda ([res : Error1])
        (
          ; Build a CPS from the exception, effectively build the "catch" block
          (o2 res)
          ; Pass the parent return channel
          ret
          ; Pass the parent exception channel
          err
        )
      )
    )
  )
)







(: safe-/
  (All [Result]
    (-> Real Real (CPS Real Symbol Result))
  )
)
(define (safe-/ x y)
  (lambda (ok err)
    (cond [(= 0 y) (err 'division-by-zero)]
          [else (ok (/ x y))])))

(: println
  (All [Error Result]
    (-> Any (CPS Void Error Result))
  )
)
(define (println x)
  (lambda (ok err)
    (displayln x)
    (ok (void))
  )
)

(define-syntax do
  (syntax-rules (<-)
    ; Only one monadic-op, return it
    [(_ mexp) mexp]
    ; A binding operation
    [(_ var : ty <- mexp rest ...) (cps-bind mexp (lambda ([var : ty]) (do rest ...)))]
    ; No binding operator, just ignore the return value
    [(_ mexp rest ...)        (cps-bind mexp (lambda (_) (do rest ...)))]))




(
  (do
    x : Real <- (safe-/ 2 1)
    (println x)
    y : Real <- (safe-/ 10 0)
    (println y)
    (try
      (safe-/ 10 0)
      ; catch
      (lambda (e)
        (do
          (println (format "Received exception: ~a" e))
          (raise "foo")
          (return (+ x y))
        )
      )
    )
  )
  ; What do you do when a normal value is returned
  (lambda (x) (display "OK: ") (displayln x) )
  ; What to do when an "exception" is raised
  (lambda (x) (display "ERROR: ") (displayln x) )
)





